//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

#include <hpgcc3.h>
#include <master_globals.h>




#define HWREG(base,off) ( (volatile unsigned int *) (((int)base+(int)off)))

#define LONG_KEYPRESSTIME (__mg->__keyb_longpresstime)
#define REPEAT_KEYTIME (__mg->__keyb_repeattime)
#define BOUNCE_KEYTIME (__mg->__keyb_debounce)

#define KF_RUNNING   1
#define KF_ALPHALOCK 2
#define KF_NOREPEAT  4




#define GPGCON ((unsigned int *)0x07a00060)
#define GPFCON ((unsigned int *)0x07a00050)
#define GPGDAT ((unsigned int *)0x07a00064)
#define EXTINT0 ((unsigned int *)0x07a00088)
#define EXTINT1 ((unsigned int *)0x07a0008c)
#define EINTMASK ((unsigned int *)0x07a000a4)
#define EINTPEND ((unsigned int *)0x07a000a8)
#define INTMSK   ((unsigned int *)0x07100008)
#define SRCPND   ((unsigned int *)0x07100000)



extern void __tmr_eventreschedule();
extern keymatrix __keyb_getmatrix();

void __keyb_postmsg(unsigned int msg)
{
	// WARNING: PROBLEMS MAY ARISE IF THE INTERRUPT SERVICE WANTS
	// TO POST A MESSAGE WHILE THE USER IS POSTING ONE.

	
__mg->__keyb_buffer[__mg->__kcurrent]=msg;
__mg->__kcurrent=(__mg->__kcurrent+1)&(KEYB_BUFFER-1);
// CHECK FOR BUFFER OVERRUN
if(__mg->__kcurrent==__mg->__kused) {
	// BUFFER OVERRUN, DROP LAST KEY
	__mg->__kcurrent=(__mg->__kcurrent-1)&(KEYB_BUFFER-1);
}
}

void keyb_postmsg(unsigned int msg)
{
	// USER VERSION, CANNOT BE INTERRUPTED BY THE INTERRUPT SERVICE
	unsigned int previous=*INTMSK;
	*INTMSK|=0x31;
	__keyb_postmsg(msg);
	*INTMSK=previous;
}


unsigned int keyb_getmsg()
{
if(__mg->__kused==__mg->__kcurrent) return 0;
unsigned int msg=__mg->__keyb_buffer[__mg->__kused];
__mg->__kused=(__mg->__kused+1)&(KEYB_BUFFER-1);
return msg;
}

// CHECK IF ANY AVAILABLE KEYSTROKES
int keyb_anymsg()
{
	if(__mg->__kused==__mg->__kcurrent) return 0;
	return 1;
}

// FLUSH KEYBOARD BUFFER
void keyb_flush()
{
	while(keyb_getmatrix()!=0LL);
	__mg->__kused=__mg->__kcurrent;
}

// NEW VERSION RETURNS THE CURRENT WORKING MATRIX INSTEAD OF
// MESSING WITH THE HARDWARE, BUT ONLY IF KEYBOARD HANDLERS WERE STARTED
keymatrix keyb_getmatrix()
{
	if(__mg->__keyflags&KF_RUNNING)	return __mg->__kmat;
	else return __keyb_getmatrix();
}

// ANALYZE CHANGES IN THE KEYBOARD STATUS AND POST MESSAGES ACCORDINGLY

void __keyb_update()
{
	keymatrix a,b;
	a=__keyb_getmatrix();
	b=a^__mg->__kmat;
	__mg->__kmat=a;

	// ANALYZE CHANGES
	if(b!=0) {
	// POST MESSAGE 	
	int key=0;
	while(b!=0) {
	if(b&1) {
	if(a&1) {
		// POST KEYDN MESSAGE
		if(__mg->__keynumber==-key) {
			// DISREGARD SPURIOUS KEYPRESS
			__mg->__kmat&=~(1LL<<key);	// CLEAR THE KEY
			__mg->__keynumber=0;
			__mg->__keycount=0;
		}
		else {
		__keyb_postmsg(KM_KEYDN + key);
		if(key<60 || ( (key==KB_ALPHA)&&(__mg->__keyplane&(SHIFT_RS|SHIFT_LS)) )) {	// TREAT SHIFT-ALPHA LIKE REGULAR KEYPRESS
		__keyb_postmsg(KM_PRESS + key + (__mg->__keyplane&SHIFT_ANY));
		__mg->__keynumber=key;
		__mg->__keycount=0;
		
		} else {
			// TODO: ADD SHIFT PLANE SWITCHING HERE
			if(key==KB_LSHIFT) {
				__mg->__keyplane|=SHIFT_LSHOLD|SHIFT_LS;
				__mg->__keyplane&=~(SHIFT_RSHOLD|SHIFT_RS | (SHIFT_RS<<16));
				__mg->__keyplane^=SHIFT_LS<<16;
			}
			if(key==KB_RSHIFT) {
				__mg->__keyplane|=SHIFT_RSHOLD|SHIFT_RS;
				__mg->__keyplane&=~(SHIFT_LSHOLD|SHIFT_LS| (SHIFT_LS<<16));
				__mg->__keyplane^=SHIFT_RS<<16;
			}
			if(key==KB_ALPHA) {
				if( __mg->__keyflags&KF_ALPHALOCK) {
					if(__mg->__keyplane&SHIFT_ALPHA)  {
						// OTHER KEYS WERE PRESSED, SO END ALPHA MODE
						__mg->__keyplane&=~((SHIFT_ALPHA<<17)|(SHIFT_ALPHA<<16)); // UNLOCK ALPHA

					} else {
							__mg->__keyplane|=SHIFT_ALPHA<<17; // LOCK ALPHA
							__mg->__keyplane&=~(SHIFT_ALPHA<<16);
						}
			
				}
				else {
				if(__mg->__keyplane&SHIFT_ALPHA)  {
					if(__mg->__keyplane&(SHIFT_ALPHA<<16)) {
						// DOUBLE ALPHA KEYPRESS
						__mg->__keyplane|=SHIFT_ALPHA<<17; // LOCK ALPHA
						__mg->__keyplane&=~(SHIFT_ALPHA<<16);
					}
					else {
						// OTHER KEYS WERE PRESSED, SO END ALPHA MODE
						__mg->__keyplane&=~((SHIFT_ALPHA<<17)|(SHIFT_ALPHA<<16)); // UNLOCK ALPHA
					}
				}
				else __mg->__keyplane^=SHIFT_ALPHA<<16;
				}
				__mg->__keyplane|=SHIFT_ALPHAHOLD|SHIFT_ALPHA;
				

			}
			if(key==KB_ON) {
				__mg->__keyplane|=SHIFT_ONHOLD;
			}
			
			__keyb_postmsg(KM_SHIFT | (__mg->__keyplane&SHIFT_ANY));
		
		}
		}
	}
	else {
		__keyb_postmsg(KM_KEYUP + key);

		if(key<60 || (__mg->__keynumber==KB_ALPHA)) {
		__mg->__keynumber=-__mg->__keynumber;
		__mg->__keycount=-BOUNCE_KEYTIME;
		__mg->__keyplane&=~((SHIFT_LS|SHIFT_RS|SHIFT_ALPHA)<<16);
		
		if(!(__mg->__keyplane& (SHIFT_LSHOLD | SHIFT_RSHOLD | SHIFT_ALPHAHOLD | SHIFT_ONHOLD))) {
			int oldkeyplane=__mg->__keyplane;
			__mg->__keyplane&=~(SHIFT_LS|SHIFT_RS|SHIFT_ALPHA); // KILL ALL SHIFT PLANES
			__mg->__keyplane|=(__mg->__keyplane>>17)&SHIFT_ALPHA; // KEEP ALPHA IF LOCKED
			if(oldkeyplane!=__mg->__keyplane)	__keyb_postmsg(KM_SHIFT | (__mg->__keyplane&SHIFT_ANY));
		}
		}
		else {
			// TODO: ADD SHIFT PLANE SWITCHING HERE
			if(key==KB_LSHIFT) {
				__mg->__keyplane&=~((SHIFT_LSHOLD|SHIFT_LS)^((__mg->__keyplane>>16)&SHIFT_LS));
				__mg->__keyplane&=~((SHIFT_ALPHA)^(((__mg->__keyplane>>16)|(__mg->__keyplane>>17))&SHIFT_ALPHA));

			}
			if(key==KB_RSHIFT) {
				__mg->__keyplane&=~((SHIFT_RSHOLD|SHIFT_RS)^((__mg->__keyplane>>16)&SHIFT_RS));
				__mg->__keyplane&=~((SHIFT_ALPHA)^(((__mg->__keyplane>>16)|(__mg->__keyplane>>17))&SHIFT_ALPHA));

			}
			if(key==KB_ALPHA) {
				__mg->__keyplane&=~((SHIFT_ALPHAHOLD|SHIFT_ALPHA)^(((__mg->__keyplane>>16)|(__mg->__keyplane>>17))&SHIFT_ALPHA));

			}
			if(key==KB_ON) {
				__mg->__keyplane&=~SHIFT_ONHOLD;
			}
			__keyb_postmsg(KM_SHIFT | (__mg->__keyplane&SHIFT_ANY));
			
			__mg->__keynumber=-key;
			__mg->__keycount=-BOUNCE_KEYTIME;
		}
		
	}
	}
	b>>=1;
	a>>=1;
	++key;
	}
	}
	
	// ANALYZE STATUS OF CURRENT KEYPRESS
	if(__mg->__kmat & (1LL<<__mg->__keynumber)) {
		// KEY STILL PRESSED, INCREASE COUNTER
		++__mg->__keycount;
		if( (__mg->__keycount>LONG_KEYPRESSTIME) && !(__mg->__keyflags&KF_NOREPEAT)) { 
			__keyb_postmsg(KM_PRESS | __mg->__keynumber | (__mg->__keyplane&SHIFT_ANY));
			__mg->__keycount=-REPEAT_KEYTIME;
		}
		if(!__mg->__keycount) {
			__keyb_postmsg(KM_PRESS | __mg->__keynumber | (__mg->__keyplane&SHIFT_ANY));
			__mg->__keycount=-REPEAT_KEYTIME;
		}
		
	}
		

	// REPEATER 
	if(__mg->__kmat==0) {
		if(__mg->__keycount>=0) { __mg->tmr_events[0].status=0; __mg->__keynumber=0; }
		else { ++__mg->__keycount; }
	} else {

		if(!(__mg->tmr_events[0].status&1)) {
			// ACTIVATE THE TIMER EVENT IF NOT ALREADY RUNNING
		if( (*HWREG(TMR_REGS,0x8))&0x100) {
		__mg->tmr_events[0].ticks=__mg->__evtmr+__mg->tmr_events[0].delay+*HWREG(TMR_REGS,0x1c)-*HWREG(TMR_REGS,0x20);
		}
		else __mg->tmr_events[0].ticks=__mg->__evtmr+__mg->tmr_events[0].delay;

		__mg->tmr_events[0].status=3;
		
		__tmr_eventreschedule();
		
		
	}
	}
	
	
	// On-C and On-A-F handling
	
	if(__mg->__kmat== ((1LL<<KB_ON) | (1LL<<KB_A)  | (1LL<<KB_F)))
	{
		// ON-A-F pressed, offer the option to stop the program
		
		throw_exception("User BREAK requested",__EX_CONT | __EX_EXIT | __EX_WARM | __EX_RESET );
		
		__mg->__kmat=0;
		
		
	}
}

void __keyb_int_handler()
{

	__keyb_update();

*EINTPEND|=0xfe00;	
	
	
}


// STATUS BUFFER REQUIRES 2 WORDS


void __keyb_init(unsigned int *keysave)
{
__mg->__keyflags=1;
__mg->__keyplane=0;	
__mg->__kused=__mg->__kcurrent=0;
__mg->__keynumber=0;
__mg->__kmat=0LL;
__mg->__keyb_repeattime=80/KEYB_SCANSPEED;
__mg->__keyb_longpresstime=1000/KEYB_SCANSPEED;
__mg->__keyb_debounce=60/KEYB_SCANSPEED;

// INITIALIZE TIMER EVENT 0

		__mg->tmr_events[0].eventhandler=__keyb_update;
		__mg->tmr_events[0].delay=(KEYB_SCANSPEED*tmr_getsysfreq())/1000;
		
		__mg->tmr_events[0].status=0;
keysave[0]=*GPGCON;
*GPGCON=0x5555AAA9;
keysave[1]=*GPFCON;
*GPFCON=(*GPFCON&0xffffc0fc)|0x2a02;
	
__irq_addhook(5,&__keyb_int_handler);
__irq_addhook(4,&__keyb_int_handler);	// SHIFTS
__irq_addhook(0,&__keyb_int_handler); // ON

*SRCPND|=1<<5;
*EXTINT0=(*EXTINT0&0xf000fff0)|0x06660006;
*EXTINT1=0x66666666;
*EINTMASK=(*EINTMASK&0xFF00018f)|0x00ff0000;
*INTMSK=*INTMSK&0xffffffce;




}


void __keyb_stop(unsigned int *keysave)
{
	
__mg->tmr_events[0].status=0;
	
// DISABLE INTERRUPTS, STATUS WILL BE FULLY RESTORED ON EXIT 
*INTMSK|=0x31;
*EINTMASK|=0xFE70;	
__irq_releasehook(5);
__irq_releasehook(4);
__irq_releasehook(0);

// RESTORE IO PORT CONFIGURATION
*GPGCON=keysave[0];
*GPFCON=keysave[1];
__mg->__keyflags&=~1;
}


void keyb_settiming(int repeat,int longpress,int debounce)
{
	__mg->__keyb_repeattime=(repeat+KEYB_SCANSPEED-1)/KEYB_SCANSPEED;
	__mg->__keyb_longpresstime=(longpress+KEYB_SCANSPEED-1)/KEYB_SCANSPEED;
	__mg->__keyb_debounce=(debounce+KEYB_SCANSPEED-1)/KEYB_SCANSPEED;
}

void keyb_setrepeat(int repeat)
{
	if(!repeat) __mg->__keyflags|=KF_NOREPEAT;
	else __mg->__keyflags&=~KF_NOREPEAT;
}

void keyb_setalphalock(int single_alpha_lock)
{
	if(single_alpha_lock) __mg->__keyflags|=KF_ALPHALOCK;
	else __mg->__keyflags&=~KF_ALPHALOCK;

}

void keyb_setshiftplane(int leftshift,int rightshift,int alpha,int alphalock)
{
	while(keyb_getmatrix()!=0LL) ;		// WAIT UNTIL NO MORE KEYS ARE PRESSED TO UPDATE SHIFT STATE
	if(leftshift) __mg->__keyplane|=SHIFT_LS|(SHIFT_LS<<16);
	else __mg->__keyplane&=~(SHIFT_LS|(SHIFT_LS<<16));
	if(rightshift) __mg->__keyplane|=SHIFT_RS|(SHIFT_RS<<16);
	else __mg->__keyplane&=~(SHIFT_RS|(SHIFT_RS<<16));
	if(alpha||alphalock) __mg->__keyplane|=SHIFT_ALPHA|(SHIFT_ALPHA<<16);
	else __mg->__keyplane&=~(SHIFT_ALPHA|(SHIFT_ALPHA<<16));
	if(alphalock) {
	__mg->__keyplane|=SHIFT_ALPHA<<17; // LOCK ALPHA
	__mg->__keyplane&=~(SHIFT_ALPHA<<16);
	}
	else {
		__mg->__keyplane&=~(SHIFT_ALPHA<<17);
	}
	keyb_postmsg(KM_SHIFT | (__mg->__keyplane&SHIFT_ANY));

}


