//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***



// $Id: sat3_real.c 544 2008-01-29 04:19:51Z claudio $

#include <hpgcc3.h>
#include <master_globals.h>

#define LN2_LN10_32  19728	// DIV BY 65536

ULONGLONG __sat3_dbl2real(double value)
{
	ULONGLONG real;

	union dblnumber {
		double dbl;
		struct { unsigned int lo,hi; };
	} number;

	if(value==0.0) real=0LL;
	else {

	number.dbl=value;
	int exp2=number.hi>>20;
	int exp10;
	int sign=exp2&0x800;
	exp2&=0x7ff;
	exp2-=1023+52;

	if(sign) number.dbl=-number.dbl;

	// GET EXPONENT IN BASE-10 TO EXTRACT
	exp10=((abs(exp2)*LN2_LN10_32)+32768)>>16;
	if(exp2>=0) {
		// USE 
		number.dbl/=i10powx(exp10);
		exp10=-exp10;
	}
	else {
		number.dbl*=i10powx(exp10);
	}
	// GET EXCEDENT EXPONENT IN BASE 2
	exp2=number.hi>>20;
	exp2&=0x7ff;
	exp2-=1023+52;

	unsigned long long mant=(((unsigned long long)((number.hi)|0x00100000)&0x001fffff)<<32);
	mant|= ((unsigned long long)number.lo);

	// NORMALIZE MANTISSA BASED ON THE EXCEDENT BASE-2 EXPONENT
	if(exp2>0) mant<<=exp2;
	else {
	mant*=10;
	++exp10;
	while(exp2<0) { mant=(mant+1)>>1; ++exp2; }	// ROUND LAST DIGIT BEFORE TRUNCATION
	}

	// HERE WE HAVE SIGN, INTEGER MANTISSA W/16 DECIMAL DIGITS (left-justified) + EXPONENT IN BASE-10

	// CONVERT MANTISSA TO 16-DIGIT BCD
	unsigned long long pow10_64[8];
	unsigned int pow10_32[8];
	unsigned int bcdhi=0,bcdlo=0,bcdmask;
	int f;
    
	pow10_32[0]=10;
	for(f=1;f<8;++f) {
		pow10_32[f]=pow10_32[f-1]*10;
	}

	for(f=0;f<8;++f) {
	pow10_64[f]=((unsigned long long)pow10_32[f])*100000000LL;
	}


	// ROUND HERE FOR 12-DIGIT REAL, EVEN ON DENORMALIZED MANTISSA
	mant+=5*pow10_32[exp10-13];

	// NORMALIZE MANTISSA IF ROUNDING UP DIDN'T CHANGE EXPONENT
	if(mant<pow10_64[7]) {
		mant*=10;
		exp10++;
	}


	for(f=7,bcdmask=0x10000000;f>=0;--f) {
		while(mant>=pow10_64[f]) {
			mant-=pow10_64[f];
			bcdhi+=bcdmask;
		}
		bcdmask>>=4;
	}

	unsigned int mant32=(unsigned int)mant;
	for(f=7,bcdmask=0x10000000;f>=0;--f) {
		while(mant32>=pow10_32[f]) {
			mant32-=pow10_32[f];
			bcdlo+=bcdmask;
		}
		bcdmask>>=4;
	}

	// CORRECT EXPONENT
	exp10=-exp10+16;
	if(exp10<0) exp10+=1000;

	// ASSEMBLE CALCULATOR REAL

	mant=(((unsigned long long)bcdhi)<<32)+(unsigned long long)bcdlo;
	mant>>=4;
	if(sign) mant|=0x9LL<<60;
	

	mant&=~0xfff;

	bcdmask=0;
	while(exp10>=100) { bcdmask+=0x100; exp10-=100; }
	while(exp10>=10)  { bcdmask+=0x10; exp10-=10; }
	bcdmask+=exp10;
	mant|=bcdmask;

    real=mant;	
	}


return real;

}







SAT_OBJECT sat3_createreal_dbl(SAT_OBJECT where,double value)
{
ULONGLONG real;

    where=sat3_alloc(where,21);
	if(where==0) return 0;
	
	real=__sat3_dbl2real(value);
    

	sat3_poke(where,SAT_DOREAL,5);
    sat3_poke(where+5,(unsigned int)real,8);
    sat3_poke(where+13,(unsigned int)(real>>32),8);

    return where;
}


#define MAX_IEEE754_EXPO 307


double __sat3_real2dbl(SAT_OBJECT realbody)
{
    
    int expo;
    double val;
        
    expo = (int) sat3_bcd2llong(realbody, 3);
    
    if (expo > 499)
        expo -= 1000;
    
    // Check for IEEE-754 exponent over/underflow...
    
    if (expo > MAX_IEEE754_EXPO)
        expo = MAX_IEEE754_EXPO;
    
    else if(expo < -MAX_IEEE754_EXPO) 
        // Treat this number as zero...
        return 0.0;

    val =  (LONGLONG)(1000 * sat3_bcd2llong (realbody + 3, 12)) * i10powx(expo-14); /* mantissa */
    
    return  (sat3_peek(realbody + 15, 1)== 9) ? -val : val;


}



double sat3_convertreal_dbl(SAT_OBJECT real)
{
if(!real) return 0.0;
if(sat3_isptr(real)) real=sat3_peek(real,5);

if(sat3_peek(real,5)!=SAT_DOREAL) return 0.0;	// RETURN 0.0 IF NOT A REAL NUMBER

real+=5;

return __sat3_real2dbl(real);

}



SAT_OBJECT sat3_createreal_int(SAT_OBJECT where,int value)
{
XREAL xr;

xr_int2xr(value,&xr);

return sat3_createreal_xreal(where,&xr);
}


int sat3_convertreal_int(SAT_OBJECT real)
{
XREAL xr;

if(sat3_convertreal_xreal(real,&xr)) return xr_xr2int(&xr);

return 0;
}


SAT_OBJECT sat3_createreal_llong(SAT_OBJECT where,LONGLONG value)
{
XREAL xr;

xr_llong2xr(value,&xr);

return sat3_createreal_xreal(where,&xr);
}


LONGLONG sat3_convertreal_llong(SAT_OBJECT real)
{
XREAL xr;

if(sat3_convertreal_xreal(real,&xr)) return xr_xr2llong(&xr);

return 0LL;
}



#define LN2_LN10_32  19728	// DIV BY 65536

ULONGLONG __sat3_xreal2real(XREAL *value)
{
XREAL xr;

	xr_base10(value,&xr);		// CONVERT EXPONENT TO BASE-10 FORMAT
	
	// HERE WE HAVE SIGN, INTEGER MANTISSA W/16 DECIMAL DIGITS (left-justified) + EXPONENT IN BASE-10

	// CONVERT MANTISSA TO 16-DIGIT BCD
	unsigned int bcdmask;
	int f;
    


	// ROUND HERE FOR 12-DIGIT REAL
	xr.Mant+=50000;
//  ROUND FOR 16-BIT EXTENDED MANTISSA
//	xr.Mant+=5;

	char tmpbuff[22];

	memset(tmpbuff,48,22);

	int ndigits=xr_getdigits10(&xr,tmpbuff);

	for(f=0,xr.Mant=0;f<15;++f) {
			xr.Mant<<=4;
			xr.Mant|=tmpbuff[f]-48;
			
	}

	// CORRECT EXPONENT
	int exp10=xr.Exp;
	
	exp10+=ndigits-1;
	if(exp10>499) {
	// OVERFLOW
	xr.Mant=0x0999999999999000LL;
	exp10=499;
	}
	else {
	if(exp10<-499) {
	// UNDERFLOW
	xr.Mant=0LL;
	exp10=0;
	xr.Sign=0;
	}
	}
	if(exp10<0) exp10+=1000;

	// ASSEMBLE CALCULATOR REAL

	if(xr.Sign) xr.Mant|=0x9LL<<60;
	
	xr.Mant&=~0xfff;

	bcdmask=0;
	while(exp10>=100) { bcdmask+=0x100; exp10-=100; }
	while(exp10>=10)  { bcdmask+=0x10; exp10-=10; }
	bcdmask+=exp10;
	xr.Mant|=bcdmask;

return xr.Mant;

}


SAT_OBJECT sat3_createreal_xreal(SAT_OBJECT where,XREAL *value)
{
ULONGLONG real;

    where=sat3_alloc(where,21);
	if(where==0) return 0;
    
	real=__sat3_xreal2real(value);

 	sat3_poke(where,SAT_DOREAL,5);
    sat3_poke(where+5,(unsigned int)real,8);
    sat3_poke(where+13,(unsigned int)(real>>32),8);

    return where;
}


#define MAX_IEEE754_EXPO 307

void __sat3_real2xreal(SAT_OBJECT realbody,XREAL *result)
{
    int expo;
            
    expo = (int) sat3_bcd2llong(realbody, 3);
    
    if (expo > 499)
        expo -= 1000;
		
	// OBTAINED BASE-10 EXPONENT
    
     result->Mant = (LONGLONG)sat3_bcd2llong (realbody + 3, 12);
	 result->Exp=expo-11;
	 result->Sign=(sat3_peek(realbody + 15, 1)== 9) ? 1 : 0;
	 
	 xr_base2(result,result);
	 
return;
}




BOOL sat3_convertreal_xreal(SAT_OBJECT real,XREAL *result)
{
if(!real) return 0;
if(sat3_isptr(real)) real=sat3_peek(real,5);

if(sat3_peek(real,5)!=SAT_DOREAL) return 0;

real+=5;
    
	__sat3_real2xreal(real,result);

	 return TRUE;
	 
}


#define LN2_LN10_32  19728	// DIV BY 65536

ULONGLONG __ullbcdadd(ULONGLONG a,ULONGLONG b)
{
ULONGLONG t1,t2,t3,t4,t5,t6;
t1=a+0x0666666666666666LL;
t2=t1+b;
t3=t1^b;
t4=t2^t3;
t5=~t4 & 0x1111111111111110LL;
t6=(t5>>2)|(t5>>3);
return t2-t6;
}



ULONGLONG __sat3_decn2real(decNumber *value)
{
	// HERE WE HAVE SIGN, INTEGER MANTISSA W/16 DECIMAL DIGITS (left-justified) + EXPONENT IN BASE-10
	XREAL xr;
	
	ULONGLONG pow10[DECDPUN+1];

	unsigned int bcdmask;
	int f,digits,shift;
	decNumberUnit *unitptr;
	int exp10;
	// CHECK OVERFLOW/UNDERFLOW FIRST
	
	if(value->exponent<-499) {
	return 0LL;
	}
	
	if(value->exponent>499) {
	return 0x0999999999999499LL;
	}
	
	// 10 POWERS TO CONVERT UNIT
    pow10[0]=1LL;
	for(f=1;f<=DECDPUN;++f) pow10[f]=pow10[f-1]*10LL;

	xr.Mant=0LL;
	// EXTRACT UNIT DIGITS
	unitptr=value->lsu+(value->digits-1)/DECDPUN;	// POINT TO LAST DIGIT UNIT
	digits=0;
	shift=0;
	while( (unitptr>=value->lsu) && (digits<16)) {
		xr.Mant*=pow10[DECDPUN];
		xr.Mant+=*unitptr;
		if(xr.Mant>pow10[DECDPUN])	digits+=DECDPUN;
		else {
			// PARTIAL NUMBER, DON'T COUNT TRAILING ZEROES
			if(*unitptr!=0) {
			int k=DECDPUN-1;
			while( *unitptr<pow10[k]) { --k; }
			digits+=k+1;
			shift+=DECDPUN-k-1;
			} else shift+=DECDPUN;
		}
		--unitptr;
	
		}
	
	// HERE WE EXTRACTED AT LEAST 16 DIGITS OUT OF THE DECNUMBER
	
	
	xr.Exp=value->exponent+(value->digits-digits);
	xr.Sign=(value->bits&DECNEG)? 1:0;
	
	while(digits<17) {
		xr.Mant*=10;
		xr.Exp--;
		++digits;
	}
	while(digits>17) {
	xr.Mant/=10;
	xr.Exp++;
	--digits;
	}
	
	// CONVERT MANTISSA TO 16-DIGIT BCD

	// ROUND HERE FOR 12-DIGIT REAL
	

	
	xr.Mant+=50000;
//  ROUND FOR 16-BIT EXTENDED MANTISSA

	char tmpbuff[22];

	memset(tmpbuff,48,22);
	int ndigits=xr_getdigits10(&xr,tmpbuff);

	for(f=1,xr.Mant=0;f<17;++f) {
			xr.Mant<<=4;
			xr.Mant|=tmpbuff[f]-48;
			
	}

	// CORRECT EXPONENT
	exp10=xr.Exp;
	
	exp10+=ndigits-3;
	if(exp10>499) {
	// OVERFLOW
	xr.Mant=0x0999999999999000LL;
	exp10=499;
	}
	else {
	if(exp10<-499) {
	// UNDERFLOW
	xr.Mant=0LL;
	exp10=0;
	xr.Sign=0;
	}
	}
	if(exp10<0) exp10+=1000;

	// ASSEMBLE CALCULATOR REAL

	if(xr.Sign) xr.Mant|=0x9LL<<60;
	
	xr.Mant&=~0xfff;

	bcdmask=0;
	while(exp10>=100) { bcdmask+=0x100; exp10-=100; }
	while(exp10>=10)  { bcdmask+=0x10; exp10-=10; }
	bcdmask+=exp10;
	xr.Mant|=bcdmask;

return xr.Mant;

}


SAT_OBJECT sat3_createreal_decn(SAT_OBJECT where,decNumber *value)
{
ULONGLONG real;

    where=sat3_alloc(where,21);
	if(where==0) return 0;
    
	real=__sat3_decn2real(value);

 	sat3_poke(where,SAT_DOREAL,5);
    sat3_poke(where+5,(unsigned int)real,8);
    sat3_poke(where+13,(unsigned int)(real>>32),8);

    return where;
}

// NOTE: result MUST HAVE BEEN ALLOCATED FOR 12 DIGITS
void __sat3_real2decn(SAT_OBJECT realbody,decNumber *result)
{
    int expo;
	decNumberUnit *unit=result->lsu;
	int f;
            
    expo = (int) sat3_bcd2llong(realbody, 3);
    
    if (expo > 499)
        expo -= 1000;
	

		result->exponent=expo-11;
		result->digits=12;
		result->bits=(sat3_peek(realbody + 15, 1)== 9) ? DECNEG : 0;
		
		realbody+=3;
		for(f=0;f<12-DECDPUN;f+=DECDPUN)
		{
			*unit++=(decNumberUnit)sat3_bcd2llong(realbody,DECDPUN);
			realbody+=DECDPUN;
			
		}
		*unit=(decNumberUnit)sat3_bcd2llong(realbody,12-f);
		for(f=12-f;f<DECDPUN;++f) *unit*=10;

return;
}




BOOL sat3_convertreal_decn(SAT_OBJECT real,decNumber *result)
{
if(!real) return 0;
if(sat3_isptr(real)) real=sat3_peek(real,5);

if(sat3_peek(real,5)!=SAT_DOREAL) return 0;

real+=5;
    
	__sat3_real2decn(real,result);

	 return TRUE;
	 
}


